#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Import the tracks table from a CSV file
"""

from sqlalchemy import create_engine
import pandas as pd
import psycopg2
import os

if os.path.exists('tracks.csv'):
    print("Please note that you need a running instance of the docker container for importing the tracks and labels tables to Postgres")
    pguser = input("User name: ")
    pgpassword = input("Password: ")
    pg = create_engine(f'postgresql://{pguser}:{pgpassword}@0.0.0.0:5555/postgres') # pg connect
    tracksimport = pd.read_csv('tracks.csv', delimiter = ',', decimal = '.')
    tracksimport.iloc[:,:91] = tracksimport.iloc[:,:91].apply(pd.to_numeric, errors='coerce', downcast='float')
    tracksimport[['trackid', 'year', 'labelid']] = tracksimport[['trackid', 'year', 'labelid']].apply(pd.to_numeric, errors='coerce', downcast='integer')
    tracksimport.to_sql('tracks', pg, if_exists='replace', method='multi', index = False, chunksize=1000)  # import to postgres
    print("Postgres table imported.")
else:
    print("Error: could not find tracks.csv")
if os.path.exists('labels/labels.csv'):
    labelsimport = pd.read_csv('labels/labels.csv', delimiter = ',', decimal = '.')
    labelsimport[['labelid', 'numtracks']] = labelsimport[['labelid', 'numtracks']].apply(pd.to_numeric, errors='coerce', downcast='integer')
    labelsimport.to_sql('labels', pg, if_exists='replace', method='multi', index = False, chunksize=1000)  # import to postgres
else:
    print("Error: could not find labels/labels.csv")
pg.dispose()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
ETL script: Processes mp3 files from the scraping directory
(mp3 tag and album cover extraction, conversion to wav, feature extraction)
then stores results in Postgres
"""

from sqlalchemy import create_engine
import time
from mutagen.easyid3 import EasyID3
from extract_features_wav import extract_features_wav_file
import pandas as pd
import logging
import os
import gc
import shutil


def gettags(filepath, track_id):
    """
    Fetches the ID3 tags from the mp3 file and extracts the cover image
    
    Returns: extracted tags and cover image link
    """
    try:        
        track = EasyID3(filepath)
    except:
        track = []
    logging.critical(track)
    if 'arranger' in track:
        coverimg = track['arranger']
    else:
        coverimg = ['']
    if 'artist' in track:
        artist = track['artist']
    else:
        artist = ['Unknown Artist']
    if 'title' in track:
        track_title = track['title']
    else:
        track_title = ['Unknown Track']
    if 'album' in track:
        album = track['album']
    else:
        album = ['']
    if 'date' in track:
        releaseyear = track['date']
        year = [int(releaseyear[0])]
    else:
        year = pd.array([None], dtype=pd.Int8Dtype())
    if 'genre' in track:
        genre = track['genre']
    else:
        genre = ['']
    if 'website' in track:
        url = track['website']
    else:
        url = ['']
    return coverimg, artist, track_title, album, year, genre, url



time.sleep(5)

pguser = os.environ.get('POSTGRES_USER')                                           # setup postgres credentials
pgpassword = os.environ.get('POSTGRES_PASSWORD')
pg = create_engine(f'postgresql://{pguser}:{pgpassword}@pg_container:5432/postgres') # pg connect

logging.basicConfig(level=logging.INFO)

while True:
    time.sleep(5)
    logging.info('Checking if scraping is underway...')
    configdata=pd.read_csv('../config/config.csv')                                  # import pipeline sraping info from config file
    scraping = configdata['scraping'].values[0]
    time.sleep(5)
    #scraping info form config.cfg -> 0: no bandcamp label under scraping or analysis; 1: scraping is underway; 2: scraper container is done with scraping
    if scraping > 0 and pg.dialect.has_table(pg, 'labels'):
        time.sleep(1)
        for file in os.listdir('../config/labels/currently_scraping'):              # first, remove any leftover wav or jpg files from previous scraping
            if file[-4:] == '.wav':
                os.remove(f'../config/labels/currently_scraping/{file}')
            elif file[-4:] == '.jpg':
                os.remove(f'../config/labels/currently_scraping/{file}')
        labeldata = pd.read_csv('../config/labels/currently_scraping/label.csv')      # get label info (label name, bandcamp (url) name, label id)
        label_name = labeldata['label_name'].values[0]
        bandcamp_name = labeldata['bandcamp_name'].values[0]                
        labeltable = pd.read_sql_table('labels', pg)
        label_id = labeltable[labeltable['bandcampname'] == bandcamp_name]['labelid'].values[0]
        track_id = 10000 * label_id + labeltable[labeltable['bandcampname'] == bandcamp_name]['numtracks'].values[0]        # generate id for next track to be analysed
        while scraping > 0:
            if pg.dialect.has_table(pg, 'tracks'):
                params = {'labelid': int(label_id)}                                 # generate a dataframe with the label tracks that are (possibly) in postgres
                label_db = pd.read_sql_query("SELECT * FROM tracks WHERE CAST(labelid AS int) = %(labelid)s;", pg, params = params)
            else:
                label_db = []
            if len(label_db) > 0:                                                   # if pre-existing label:
                label_exists = 1
                if label_db.shape[0] != labeltable[labeltable['labelid'] == label_id]['numtracks'].values[0]: # set correct number of tracks in case of previous error / crash during scraping
                    labeltable.loc[labeltable[labeltable['labelid'] == label_id].index.values[0],'numtracks'] = label_db.shape[0]
                    labeltable.to_sql('labels', pg, if_exists='replace', method='multi', index = False, chunksize=1000)
                    track_id = 10000 * label_id + label_db.shape[0]
                    logging.info(f'Track num readjusted to {label_db.shape[0]}.')
            else:
                label_exists = 0
            files = sorted(os.listdir('../config/labels/currently_scraping'))       # get list of already scraped files, so the two containers can work simultaneously
            data = []
            files_processed = 1     # value will be changed to 0 once a valid file is found for analysis
            for file in files:
                if file[-4:] == '.mp3':
                    filepath = f'../config/labels/currently_scraping/{file}'
                    total, used, free = shutil.disk_usage("../config")
                    if label_exists and file in label_db['filename'].values:        # perhaps file is already in the database:
                        logging.info(f'Skipping file already in the database: {file}')
                        os.remove(filepath)                                         # clean up the mp3 file
                    elif free < 100000000:                                          # perhaps running out of disk space:
                        logging.warning(f'Available space very low (under 100MB). Skipping {file}.')
                        os.remove(filepath)                                         # clean up the mp3 file
                        time.sleep(5)
                    else:
                        logging.info(f'Processing: {file}')
                        files_processed = 0
                        track_id += 1

                        coverimg, artist, track_title, album, year, genre, url = gettags(filepath, track_id)
                        filename = file
                        dftrack = []

                        try:                        
                            os.system(f'ffmpeg -i "{filepath}" -acodec pcm_s16le -ar 22050 -ac 1 -nostats -loglevel 0 -y "{filepath}.wav"') # convert to wav
                            dftrack = extract_features_wav_file(f'{filepath}.wav')   # extract the audio features
                            os.remove(f'{filepath}.wav')                             # clean up the wav file
                        except Exception as e:
                            logging.warning(f'Could not process file: {file}. Error: {str(e)}')
                            track_id -= 1

                        if os.path.exists(filepath):
                            os.remove(filepath)                                         # clean up the mp3 file
                        gc.collect()
                        
                        if len(dftrack) > 0:                                        # update postgres and number of tracks in labels CSV
                            dftrack['trackid'], dftrack['artist'], dftrack['year'], dftrack['track'], dftrack['album'], dftrack['tags'], dftrack['labelid'], dftrack['labelname'], dftrack['filename'], dftrack['url'], dftrack['bclabel'], dftrack['coverimg'] = track_id, artist[0], year[0], track_title[0], album[0], genre[0], int(label_id), label_name, file, url[0], bandcamp_name, coverimg
                            dftrack.iloc[0,:91] = dftrack.iloc[0,:91].apply(pd.to_numeric, downcast='float')
                            dftrack[['trackid', 'labelid']] = dftrack[['trackid', 'labelid']].apply(pd.to_numeric, downcast='integer')
                            dftrack.to_sql('tracks', pg, if_exists='append', method='multi', index = False, chunksize=1000)  # export to postgres
                            labeltable.loc[labeltable[labeltable['labelid'] == label_id].index.values[0],'numtracks'] = track_id % 10000
                            labeltable.to_sql('labels', pg, if_exists='replace', method='multi', index = False, chunksize=1000)

            if files_processed == 1:                                                  # if no valid files found, perhaps the scraping is over
                if scraping == 2:           
                    logging.info(f'Processing of label {label_name} complete!')
                    scraping = 0
                else:       # scraping might be over, but only after doing an extra loop of file search (perhaps another file was downloaded while analysing)
                    logging.info('No new files found in scraping directory, giving it another go...')
                    time.sleep(5)
                    configdata=pd.read_csv('../config/config.csv')                    # check pipeline scraping status
                    scraping = configdata['scraping'].values[0]                       # if the value is 2 (scraping container is finished with job), the loop will run one more time to check if new files were downloaded during the analysis
        labeltable.loc[labeltable[labeltable['labelid'] == label_id].index.values[0],'ready'] = True
        labeltable.to_sql('labels', pg, if_exists='replace', method='multi', index = False, chunksize=1000)        
        configdata['scraping'].values[0] = 0
        configdata.to_csv('../config/config.csv', index=False)                        # inform pipeline that scraping and analysis is over; set info in postgres labels table too



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

Predicts musical subgenres of electronic dance music tracks in CSV file received through command-line argument
Follows the methodology of Caparrini et al. (https://doi.org/10.1080/09298215.2020.1761399)
Uses 92 features extracted with pyAudioAnalysis and Essentia
These features should be located in the first 92 columns of the CSV in the default order of extraction

"""

import pickle
import pandas as pd
import argparse
import os.path

parser = argparse.ArgumentParser(description='Predicts musical genre of electronic dance music tracks based on 92 musical features, using a model trained on Beatport Top100 lists (cf. Caparrini et al. 2020)')
parser.add_argument('-f', '--file', help='CSV file with 92 pyAudioAnalysis and Essentia features', type = str)

args = parser.parse_args()
csv_file = args.file

if csv_file == None:
    print ('Please provde the list of track featuers in a CSV file. Command line argument: -f {filename} or --file {filename}')
else:
    if os.path.isfile(csv_file):
        df = pd.read_csv(csv_file)
        clf = pickle.load(open("../config/beatport_classifier.sav", 'rb'))
        features_table = df.iloc[:,:92]
        df = df.assign(Predicted_Genre=pd.Series(clf.predict(features_table.to_numpy())))
        resultsfile = csv_file[:-4] + '_predictions' + csv_file[-4:]
        df.to_csv(resultsfile, index=False)
        print(f'Results saved into: {resultsfile}')
    else:
        print(f'File not found: {csv_file}')


#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
Bandcamp Scraper
Much of this code is borrowed from Rich Jones' SoundScrape
Miserlou / SoundScrape - https://github.com/Miserlou/SoundScrape
"""

import demjson
import html
import os
import re
import requests
import sys
import logging
import time
import pandas as pd
import shutil

from clint.textui import colored, puts, progress
from datetime import datetime
from mutagen.mp3 import MP3, EasyMP3
from mutagen.id3 import APIC, WXXX
from mutagen.id3 import ID3 as OldID3
from subprocess import Popen, PIPE
from os.path import dirname, exists, join
from os import access, mkdir, W_OK

from sqlalchemy import create_engine

if sys.version_info.minor < 4:
    html_unescape = html.parser.HTMLParser().unescape
else:
    html_unescape = html.unescape



####################################################################
# Bandcamp
####################################################################


# Largely borrowed from Ronier's bandcampscrape
def scrape_bandcamp_url(labelid, url, sizelimit, alreadyscraped, num_tracks=sys.maxsize, folders=False, custom_path=''):
    """
    Pull out artist and track info from a Bandcamp URL.

    Returns:
        list: filenames to open
        
    For the current purpose, it scrapes everything into the same folder and attaches a labelid to the flies
    Files will be deleted after feature extractions

    """

    filenames = []
    album_data = get_bandcamp_metadata(url)
    # If it's a list, we're dealing with a list of Album URLs,
    # so we call the scrape_bandcamp_url() method for each one
    if type(album_data) is list:
        for album_url in album_data:
            filenames.append(
                scrape_bandcamp_url(
                    labelid, album_url, sizelimit, alreadyscraped, num_tracks, folders, custom_path
                )
            )
        return filenames

    artist = album_data.get("artist")
    album_name = album_data.get("album_title")

    if folders:
        if album_name:
            directory = artist + " - " + album_name
        else:
            directory = artist
        directory = sanitize_filename(directory)
        directory = join(custom_path, directory)
        if not exists(directory):
            mkdir(directory)

    for i, track in enumerate(album_data["trackinfo"]):

        if i > num_tracks - 1:
            continue

        try:
            total, used, free = shutil.disk_usage("../config")
            tag_tracklink = url[0:url.find("bandcamp.com") + 12] + track["title_link"]            
            if free < 200000000:
                while free < 200000000:
                    logging.warning(f'Running out of space (under 200MB). Trying to download {tag_tracklink} again in a minute.')
                    total, used, free = shutil.disk_usage("../config")
                    time.sleep(60)
            if tag_tracklink in alreadyscraped:
                logging.info(f'Track {tag_tracklink} is already in the database. Skipping.')
            else:

                track_name = track["title"]
                if track["track_num"]:
                    track_number = str(track["track_num"]).zfill(2)
                else:
                    track_number = None
                if track_number and folders:
                    track_filename = '%s - %s.mp3' % (track_number, track_name)
                else:
                    track_filename = '%s.mp3' % (track_name)

                track_filename = sanitize_filename(track_filename)

                if folders:
                    path = join(directory, track_filename)
                else:
                    path = join(custom_path, str(labelid) + '_' + sanitize_filename(artist) + ' - ' + track_filename)

                if exists(path):
                    puts_safe(colored.yellow("Track already downloaded: ") + colored.white(track_name))
                    continue

                if not track['file']:
                    puts_safe(colored.yellow("Track unavailble for scraping: ") + colored.white(track_name))
                    continue

                puts_safe(colored.green("Downloading") + colored.white(": " + track_name))
                path = download_file(track['file']['mp3-128'], path, sizelimit)

                if path == "too_large":
                    logging.info(f'File exceeds the size limit of {sizelimit} bytes. Skipping.')
                    raise Exception('File exceeds size limit.')

                album_year = album_data['album_release_date']
                if album_year:
                    album_year = datetime.strptime(album_year, "%d %b %Y %H:%M:%S GMT").year

                tag_file(path,
                     artist,
                     track_name,
                     album=album_name,
                     year=album_year,
                     genre=album_data['genre'],
                     artwork_url=album_data['artFullsizeUrl'],
                     track_number=track_number,
                     url=tag_tracklink,
                     )

                filenames.append(path)

                logging.info(f'Download complete: {artist} - {track_name}')

        except Exception as e:
            logging.warning("Problem downloading")
            print(e)

    return filenames




def extract_embedded_json_from_attribute(request, attribute, debug=False):
    """
    Extract JSON object embedded in an element's attribute value.

    The JSON is "sloppy". The native python JSON parser often can't deal,
    so we use the more tolerant demjson instead.

    Args:
        request (obj:`requests.Response`): HTTP GET response from which to extract
        attribute (str): name of the attribute holding the desired JSON object
        debug (bool, optional): whether to print debug messages

    Returns:
        The embedded JSON object as a dict, or None if extraction failed
    """
    try:
        embed = request.text.split('{}="'.format(attribute))[1]
        embed = html_unescape(
            embed.split('"')[0]
        )
        output = demjson.decode(embed)
        if debug:
            print(
                'extracted JSON: '
                + demjson.encode(
                    output,
                    compactly=False,
                    indent_amount=2,
                )
            )
    except Exception as e:
        output = None
        if debug:
            print(e)
    return output


def get_bandcamp_metadata(url):
    """
    Read information from Bandcamp embedded JavaScript object notation.
    The method may return a list of URLs (indicating this is probably a "main" page which links to one or more albums),
    or a JSON if we can already parse album/track info from the given url.
    """
    request = requests.get(url)
    output = {}
    try:
        for attr in ['data-tralbum', 'data-embed']:
            output.update(
                extract_embedded_json_from_attribute(
                    request, attr
                )
            )
    # if the JSON parser failed, we should consider it's a "/music" page,
    # so we generate a list of albums/tracks and return it immediately
    except Exception as e:
        regex_all_albums = r'<a href="(.*/(?:album|track)/[^>]+)">'
        all_albums = re.findall(regex_all_albums, request.text, re.MULTILINE)
        album_url_list = list()
        for album in all_albums:
            album_url = album
            if "?" in album:
                album_url = album[0:album.find("?")]
            if album[0:4] != "http":
                album_url = re.sub(r'music/?$', '', url) + album
            album_url_list.append(album_url)
        return album_url_list
    # if the JSON parser was successful, use a regex to get all tags
    # from this album/track, join them and set it as the "genre"
    regex_tags = r'<a class="tag" href[^>]+>([^<]+)</a>'
    tags = re.findall(regex_tags, request.text, re.MULTILINE)
    # make sure we treat integers correctly with join()
    # according to http://stackoverflow.com/a/7323861
    # (very unlikely, but better safe than sorry!)
    output['genre'] = '|'.join(s for s in tags)

    try:
        artUrl = request.text.split("\"tralbumArt\">")[1].split("\">")[0].split("href=\"")[1]
        output['artFullsizeUrl'] = artUrl
    except:
        puts_safe(colored.red("Couldn't get full artwork") + "")
        output['artFullsizeUrl'] = None

    return output



####################################################################
# File Utility
####################################################################


def download_file(url, path, sizelimit, session=None, params=None):
    """
    Download an individual file.
    """

    if url[0:2] == '//':
        url = 'https://' + url[2:]

    # Use a temporary file so that we don't import incomplete files.
    tmp_path = path + '.tmp'

    if session and params:
        r = session.get( url, params=params, stream=True )
    elif session and not params:
        r = session.get( url, stream=True )
    else:
        r = requests.get(url, stream=True)

    logging.info(f'Downloading {int(r.headers.get("content-length", 0))} bytes.')

    if int(r.headers.get('content-length', 0)) < int(sizelimit):
        with open(tmp_path, 'wb') as f:
            total_length = int(r.headers.get('content-length', 0))
            for chunk in progress.bar(r.iter_content(chunk_size=1024), expected_size=(total_length / 1024) + 1):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)
                    f.flush()
        os.rename(tmp_path, path)
        return path
    else:
        return "too_large"


def tag_file(filename, artist, title, year=None, genre=None, artwork_url=None, album=None, track_number=None, url=None, comment=None):
    """
    Attempt to put ID3 tags on a file.

    Args:
        artist (str):
        title (str):
        year (int):
        genre (str):
        artwork_url (str):
        album (str):
        track_number (str):
        filename (str):
        url (str):
    """

    try:
        audio = EasyMP3(filename)
        audio.tags = None
        audio["artist"] = artist
        audio["title"] = title
        if year:
            audio["date"] = str(year)
        if album:
            audio["album"] = album
        if track_number:
            audio["tracknumber"] = track_number
        if genre:
            audio["genre"] = genre
        if url:
            audio["website"] = url
        if artwork_url:
            artwork_url = artwork_url.replace('https', 'http')
            audio["arranger"] = artwork_url
        audio.save()


        # because there is software that doesn't seem to use WOAR we save url tag again as WXXX
        if url:
            audio = MP3(filename, ID3=OldID3)
            audio.tags.add( WXXX( encoding=3, url=url ) )
            audio.save()

        return True

    except Exception as e:
        puts(colored.red("Problem tagging file: ") + colored.white("Is this file a WAV?"))
        return False

def open_files(filenames):
    """
    Call the system 'open' command on a file.
    """
    command = ['open'] + filenames
    process = Popen(command, stdout=PIPE, stderr=PIPE)
    stdout, stderr = process.communicate()


def sanitize_filename(filename):
    """
    Make sure filenames are valid paths.

    Returns:
        str:
    """
    sanitized_filename = re.sub(r'[/\\:*?"<>|]', '-', filename)
    sanitized_filename = sanitized_filename.replace('&', 'and')
    sanitized_filename = sanitized_filename.replace('"', '')
    sanitized_filename = sanitized_filename.replace("'", '')
    sanitized_filename = sanitized_filename.replace("/", '')
    sanitized_filename = sanitized_filename.replace("\\", '')

    # Annoying.
    if sanitized_filename[0] == '.':
        sanitized_filename = u'dot' + sanitized_filename[1:]

    return sanitized_filename

def puts_safe(text):
    if sys.platform == "win32":
        if sys.version_info < (3,0,0):
            puts(text)
        else:
            puts(text.encode(sys.stdout.encoding, errors='replace').decode())
    else:
        puts(text)



####################################################################
# Bandcamp scraper for the label classifier app
####################################################################



def get_label_tracklist(bandcampname,fullname):
    '''
    fetches or generates label ID and updates the labels postgres and CSV table if needed
    '''

    alreadyscraped = []
    pguser = os.environ.get('POSTGRES_USER')
    pgpassword = os.environ.get('POSTGRES_PASSWORD')
    pg = create_engine(f'postgresql://{pguser}:{pgpassword}@pg_container:5432/postgres')  # pg connect    
    if pg.dialect.has_table(pg, 'labels'):
        
        labeltable = pd.read_sql_table('labels', pg)
        if len(labeltable[labeltable['bandcampname'] == bandcampname]) == 0:            # if new label in list:
            labelid = labeltable['labelid'].max() + 1
            labeltable = labeltable.append(pd.DataFrame({'labelid' : [labelid], 'bandcampname' : [bandcampname], 'labelname' : [fullname], 'numtracks' : [0]}), ignore_index = True)
        else:                                                                           # if pre-existing label:
            labelid = labeltable[labeltable['bandcampname'] == bandcampname]['labelid'].values[0]
            if pg.dialect.has_table(pg, 'tracks'):
                params = {'labelid': int(labelid)}                                      # check label tracks already in postgres
                label_db = pd.read_sql_query("SELECT * FROM tracks WHERE CAST(labelid AS int) = %(labelid)s;", pg, params = params)            
                alreadyscraped = label_db['url'].tolist()
    else:                                                                               # if this is te first label being scraped:
        labelid = 1
        labeltable = pd.DataFrame({'labelid' : [1], 'bandcampname' : [bandcampname], 'labelname' : [fullname], 'numtracks' : [0], 'ready' : [False]})
    
    labeltable['ready'] = True
    labeltable.loc[labeltable[labeltable['labelid'] == labelid].index.values[0],'ready'] = False
    labeltable.to_sql('labels', pg, if_exists='replace', method='multi', index = False, chunksize=1000)  # export label table to postgres too
    pg.dispose()

    return labelid, alreadyscraped



logging.basicConfig(level=logging.INFO)
while True:
    time.sleep(5)
    pguser = os.environ.get('POSTGRES_USER')
    pgpassword = os.environ.get('POSTGRES_PASSWORD')
    pg = create_engine(f'postgresql://{pguser}:{pgpassword}@pg_container:5432/postgres')  # pg connect                
    logging.info('Checking if a new label has been assigned for scraping...')
    configdata=pd.read_csv('../config/config.csv')                          # import config settings
    bc_label = configdata['bclabel'].values[0]
    sizelimit = configdata['sizelimit'].values[0]

    if bc_label != '_none_':                                                # if label assigned for scraping:
        logging.info(f'Checking the Bandcamp page of {bc_label}...')
        configdata['scraping'].values[0] = 1
        configdata.to_csv('../config/config.csv', index=False)              # notify pipeline that scraping is underway through config file
        label_url = 'https://' + bc_label + '.bandcamp.com/music'
        request = requests.get(label_url)
        regex_title = r'<meta property="og:site_name" content=".*">'
        labelname = re.findall(regex_title, request.text)[0]
        labelname = labelname[39:-2]                                        # get label name
        labelid, alreadyscraped = get_label_tracklist(bc_label, labelname)  # get label id and list of tracks already scraped
        for file in os.listdir('../config/labels/currently_scraping'):      # remove any leftover mp3 files from scraping folder
            if file[-4:] == '.mp3':
                os.remove(f'../config/labels/currently_scraping/{file}')
        newlabelinfo = pd.DataFrame ({'label_name' : [labelname], 'bandcamp_name' : [bc_label]})
        newlabelinfo.to_csv('../config/labels/currently_scraping/label.csv',index=False)        # add label details to pipeline, to be fetched by the etl container
        bc_url = 'https://' + bc_label + '.bandcamp.com/music'
        filenames = scrape_bandcamp_url(labelid, bc_url, sizelimit, alreadyscraped, custom_path = '../config/labels/currently_scraping/')  # scrape bandcamp files

        configdata['bclabel'].values[0] = "_none_"                          # scraping complete
        configdata['scraping'].values[0] = 2
        configdata.to_csv('../config/config.csv', index=False)              # notify pipeline that scraping is complete
    
    time.sleep(5)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Extracts 92 pyAudio and Essentia features from a .wav (22050 Hz mono) file
Based on pyGenreClf / A. Caparrini
https://github.com/Caparrini/pyGenreClf
"""


from pyAudioAnalysis import audioBasicIO
from pyAudioAnalysis.MidTermFeatures import mid_feature_extraction
from pyAudioAnalysis.MidTermFeatures import beat_extraction
import pandas as pd
import numpy as np

import essentia.standard as es
import time


def file_extract_essentia(audio_file):
    """
    Essentia feature extraction / tailored code from pyGenreClf
    
    Returns: features
    """
    rhythm_feats = ["rhythm.bpm",
                    "rhythm.bpm_histogram_first_peak_bpm",
                    "rhythm.bpm_histogram_first_peak_weight",
                    "rhythm.bpm_histogram_second_peak_bpm",
                    "rhythm.bpm_histogram_second_peak_spread",
                    "rhythm.bpm_histogram_second_peak_weight",
                    "rhythm.danceability",
                    "rhythm.beats_loudness.mean",
                    "rhythm.beats_loudness.stdev",
                    "rhythm.onset_rate",
                    "rhythm.beats_loudness_band_ratio.mean",
                    "rhythm.beats_loudness_band_ratio.stdev"
                    ]

    features_total, features_frames = es.MusicExtractor()(audio_file)

    features = []
    for i in range(0,len(rhythm_feats)-2):
        features.append(features_total[rhythm_feats[i]])

    for i in range(len(rhythm_feats)-2,len(rhythm_feats)):
        bands = features_total[rhythm_feats[i]]
        for j in range(0,len(bands)):
            features.append(bands[j])

    return np.array(features)



def feature_extract_pyAudio(file_path, mid_window, mid_step, short_window, short_step):
    """
    pyAudio feature extraction / tailored code from pyGenreClf
    
    Returns: features
    """
    mid_term_features = np.array([])
    process_times = []
    mid_feature_names = []   
    sampling_rate, signal = audioBasicIO.read_audio_file(file_path)
    t1 = time.time()        
    signal = audioBasicIO.stereo_to_mono(signal)
    if signal.shape[0] < float(sampling_rate)/5:
        print("  (AUDIO FILE TOO SMALL - SKIPPING)")
    else:
        mid_features, short_features, mid_feature_names = \
            mid_feature_extraction(signal, sampling_rate,
                                   round(mid_window * sampling_rate),
                                   round(mid_step * sampling_rate),
                                   round(sampling_rate * short_window),
                                   round(sampling_rate * short_step))
        beat, beat_conf = beat_extraction(short_features, short_step)
        
        mid_features = np.transpose(mid_features)
        mid_features = mid_features.mean(axis=0)
        # long term averaging of mid-term statistics
        if (not np.isnan(mid_features).any()) and \
                (not np.isinf(mid_features).any()):
            mid_features = np.append(mid_features, beat)
            mid_features = np.append(mid_features, beat_conf)
            if len(mid_term_features) == 0:
                # append feature vector
                mid_term_features = mid_features
            else:
                mid_term_features = np.vstack((mid_term_features, mid_features))
            t2 = time.time()
            duration = float(len(signal)) / sampling_rate
            process_times.append((t2 - t1) / duration)
        if len(process_times) > 0:
            print("Feature extraction complexity ratio: "
                  "{0:.1f} x realtime".format((1.0 / np.mean(np.array(process_times)))))
    return mid_term_features, mid_feature_names



def extract_features_wav_file(file_path):    
    """
    Extracts the 92 audio features
    
    Returns: 1-row dataframe with features
    """
    data = []
    feature_values, feature_names = feature_extract_pyAudio(file_path, 1, 1, 0.05, 0.05)
    feature_names = feature_names + ['BPM','BPMConf']
    feature_names = feature_names + ['71-BPMessentia','72-bpm_histogram_first_peak_bpm','73-bpm_histogram_first_peak_weight','74-bpm_histogram_second_peak_bpm','75-bpm_histogram_second_peak_spread','76-bpm_histogram_second_peak_weight','77-danceability','78-beats_loudness.mean','79-beats_loudness.stdev','80-onset_rate','81-beats_loudness_band_ratio.mean1','82-beats_loudness_band_ratio.mean2','83-beats_loudness_band_ratio.mean3','84-beats_loudness_band_ratio.mean4','85-beats_loudness_band_ratio.mean5','86-beats_loudness_band_ratio.mean6','87-beats_loudness_band_ratio.stdev1','88-beats_loudness_band_ratio.stdev2','89-beats_loudness_band_ratio.stdev3','90-beats_loudness_band_ratio.stdev4','91-beats_loudness_band_ratio.stdev5','92-beats_loudness_band_ratio.stdev6']
    rhythm_feats = file_extract_essentia(file_path)
    data.append(feature_values.tolist() + rhythm_feats.tolist())
    df = pd.DataFrame(np.array(data), columns=feature_names)
    df.drop([col for col in df.columns if 'delta' in col],axis=1,inplace=True)          # drop delta values
    df.columns = ['1-ZCRm', '2-Energym', '3-EnergyEntropym', '4-SpectralCentroidm', '5-SpectralSpreadm', '6-SpectralEntropym', '7-SpectralFluxm', '8-SpectralRolloffm', '9-MFCCs1m', '10-MFCCs2m', '11-MFCCs3m', '12-MFCCs4m', '13-MFCCs5m', '14-MFCCs6m', '15-MFCCs7m', '16-MFCCs8m', '17-MFCCs9m', '18-MFCCs10m', '19-MFCCs11m', '20-MFCCs12m', '21-MFCCs13m', '22-ChromaVector1m', '23-ChromaVector2m', '24-ChromaVector3m', '25-ChromaVector4m', '26-ChromaVector5m', '27-ChromaVector6m', '28-ChromaVector7m', '29-ChromaVector8m', '30-ChromaVector9m', '31-ChromaVector10m', '32-ChromaVector11m', '33-ChromaVector12m', '34-ChromaDeviationm', '35-ZCRstd', '36-Energystd', '37-EnergyEntropystd', '38-SpectralCentroidstd', '39-SpectralSpreadstd', '40-SpectralEntropystd', '41-SpectralFluxstd', '42-SpectralRolloffstd', '43-MFCCs1std', '44-MFCCs2std', '45-MFCCs3std', '46-MFCCs4std', '47-MFCCs5std', '48-MFCCs6std', '49-MFCCs7std', '50-MFCCs8std', '51-MFCCs9std', '52-MFCCs10std', '53-MFCCs11std', '54-MFCCs12std', '55-MFCCs13std', '56-ChromaVector1std', '57-ChromaVector2std', '58-ChromaVector3std', '59-ChromaVector4std', '60-ChromaVector5std', '61-ChromaVector6std', '62-ChromaVector7std', '63-ChromaVector8std', '64-ChromaVector9std', '65-ChromaVector10std', '66-ChromaVector11std', '67-ChromaVector12std', '68-ChromaDeviationstd', '69-BPM', '70-BPMconf', '71-BPMessentia','72-bpm_histogram_first_peak_bpm','73-bpm_histogram_first_peak_weight','74-bpm_histogram_second_peak_bpm','75-bpm_histogram_second_peak_spread','76-bpm_histogram_second_peak_weight','77-danceability','78-beats_loudness.mean','79-beats_loudness.stdev','80-onset_rate','81-beats_loudness_band_ratio.mean1','82-beats_loudness_band_ratio.mean2','83-beats_loudness_band_ratio.mean3','84-beats_loudness_band_ratio.mean4','85-beats_loudness_band_ratio.mean5','86-beats_loudness_band_ratio.mean6','87-beats_loudness_band_ratio.stdev1','88-beats_loudness_band_ratio.stdev2','89-beats_loudness_band_ratio.stdev3','90-beats_loudness_band_ratio.stdev4','91-beats_loudness_band_ratio.stdev5','92-beats_loudness_band_ratio.stdev6']
    # the final column names replicate the column names of Caparrini's Beatport datasets; see https://www.kaggle.com/caparrini/electronic-music-features-201802-beatporttop100
    return df




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Flask application for the Bandcamp label classificator / admin interface

"""

from flask import Flask, render_template, request, send_from_directory, jsonify, abort
import logging
import pandas as pd
from clusterer import process_label
from sqlalchemy import create_engine
import psycopg2
import os
import requests
import re
from waitress import serve


app = Flask(__name__)


@app.route('/index')
@app.route('/')
def index():                                                            # main page for clusterisation
    global labels
    pguser = os.environ.get('POSTGRES_USER')
    pgpassword = os.environ.get('POSTGRES_PASSWORD')
    pg = create_engine(f'postgresql://{pguser}:{pgpassword}@pg_container:5432/postgres')  # pg connect    
    if pg.dialect.has_table(pg, 'labels'):
        labeltable = pd.read_sql_table('labels', pg)                    # check available labels
        pg.dispose()
        labels = labeltable[(labeltable['numtracks'] > 49) & (labeltable['ready'])]['bandcampname'].tolist()
        labels.sort()
        return render_template("index.html", labels = labels)           # render index.html with selectable list of labels
    else:
        pg.dispose()
        return render_template("index.html", labels = [])               # nothing to classify yet



@app.route('/warning_message', methods=['POST'])
def warning_message():
    global processing
    if processing == True:
        return jsonify(message='Processing, please wait!')
    else:
        return jsonify(message='')


@app.route('/classifier', methods=['GET'])
def classifier():
    global processing
    processing = True
    global labels
    usrinput = dict(request.args)
    if 'labels' in usrinput:
        labeltitle = usrinput['labels']
        clusternum = usrinput['clusternum']
        if clusternum == 'Auto':
            clusternum = 0
            logging.info(f'Distributing {labeltitle} into automatically determined number of clusters...')
        else:
            logging.info(f'Distributing {labeltitle} into {clusternum} clusters...')
        process_label(labeltitle, clusternum)
        processing = False
        try:
            return send_from_directory(directory='../config/labels',filename=f'{labeltitle}{clusternum}.pdf',mimetype='application/pdf', as_attachment=True)
        except FileNotFoundError:
            abort(404)
    else:
        return render_template("index.html", labels = [])               # nothing to classify yet
            

@app.route('/addlibrary', methods=['GET'])                              # add new library for processing (admin interface)
def addlibrary():
    usrinput = dict(request.args)
    if len(usrinput['libname']) > 0:
        libname = usrinput['libname']
        urlrequest = requests.get(f'https://{libname}.bandcamp.com/music')
        try:
            regex_title = r'<meta property="og:site_name" content=".*">'
            labelname = re.findall(regex_title, urlrequest.text)[0]     # this will generate an exception if the url does not point to a bandcamp music page
            labelname = labelname[39:-2]
            configdata=pd.read_csv('../config/config.csv')              # import sraping info from config file
            if configdata['scraping'].values[0] > 0:                    # if another scraping/analysis is underway, try to estimate the number of files left
                labeldata=pd.read_csv('../config/labels/currently_scraping/label.csv')
                if configdata['scraping'].values[0] == 2:
                    filesleftnum = str(len([x for x in os.listdir('../config/labels/currently_scraping') if (x.endswith('.mp3'))]))
                else:
                    filesleftnum = "unknown"
                return render_template("admin.html", warning_message = f"Submission not possible at this time.. currently analysing the Bandcamp releases of {labeldata['label_name'].values[0]}. Number of files left: {filesleftnum}, please try again later.", disabled = "disabled")
            else:
                configdata['bclabel'].values[0] = libname               # set pipeline config file to start processing the files
                configdata.to_csv('../config/config.csv', index=False)
                return render_template("admin.html", warning_message = f"Success! The analysis of {labelname} tracks on Bandcamp is underway.\nThis may take several (or many) hours. Number of files left: unknown yet, please check back later.", disabled = "disabled")
        except:
            return render_template("admin.html", warning_message = f"https://{libname}.bandcamp.com is not a valid bandcamp label/library page.")
    else:
        return render_template("admin.html", warning_message = "Please enter a valid bandcamp label/library page.")        


@app.route('/admin', methods=['GET'])                                   # admin page: needs to show if new label submission is possible or not (when another label is being processed already)
def admin():
    configdata=pd.read_csv('../config/config.csv')                      # import sraping info from config file
    if configdata['scraping'].values[0] > 0:
        labeldata=pd.read_csv('../config/labels/currently_scraping/label.csv')
        if configdata['scraping'].values[0] == 2:                       # estimate number of files left for processing:
            filesleftnum = str(len([x for x in os.listdir('../config/labels/currently_scraping') if (x.endswith('.mp3'))]))
        else:
            filesleftnum = "unknown yet"
        return render_template('admin.html', warning_message = f"Submission not possible at this time.. currently analysing the Bandcamp releases of {labeldata['label_name'].values[0]}. Number of files left: {filesleftnum}, please try again later.", disabled = "disabled")
    else:
        return render_template('admin.html', warning_message = "", disabled = "")


@app.route('/gettracks', methods=['GET'])                               # get tracks table (admin interface)
def gettracks():
    pguser = os.environ.get('POSTGRES_USER')                            # setup postgres connection
    pgpassword = os.environ.get('POSTGRES_PASSWORD')
    pg = create_engine(f'postgresql://{pguser}:{pgpassword}@pg_container:5432/postgres')
    if pg.dialect.has_table(pg, 'tracks'):
        tracks = pd.read_sql_table('tracks', pg, columns=['trackid','artist','year','track','album','tags','labelid','labelname','url','bclabel'])
        pg.dispose()
        tracks.to_csv('tracks.csv', index=False)                        # fetch tracks table
        return send_from_directory(directory='.',filename='tracks.csv',mimetype='text/csv', as_attachment=True)
    else:
        pg.dispose()
        return render_template('admin.html', warning_message = "Tracks database not found.", disabled = "")        


@app.route('/getlabels', methods=['GET'])                               # get label shortlist (admin interface)
def getlabels():
    pguser = os.environ.get('POSTGRES_USER')
    pgpassword = os.environ.get('POSTGRES_PASSWORD')
    pg = create_engine(f'postgresql://{pguser}:{pgpassword}@pg_container:5432/postgres')  # pg connect    
    if pg.dialect.has_table(pg, 'labels'):
        labeltable = pd.read_sql_table('labels', pg)
        pg.dispose()
        labeltable.to_csv('labels.csv', index=False)                    # fetch labels table
        return send_from_directory(directory='.',filename='labels.csv',mimetype='text/csv', as_attachment=True)
    else:
        pg.dispose()
        return render_template('admin.html', warning_message = "No Bandcamp labels/libraries analysed yet", disabled = "")

if __name__ == '__main__':
    global processing
    processing = False
    logging.basicConfig(level=logging.INFO)    
    app.config['SECRET_KEY'] = os.environ.get('FLASK_KEY')
    serve(app, host="0.0.0.0", threads=10, port=8000)



"""Console script for Bandcamp Dance Librarian."""

import pandas as pd
import os
import argparse
import sys
import subprocess
import psycopg2
import time
    
from sqlalchemy import create_engine


def import_tracks(postgres_pwd):
    """
    Imports tracks and labels tables to Postgres database

    """
    if os.path.exists('config/tracks.csv') and os.path.exists('config/labels/labels.csv'):
        print('Importing tables to Postgres database..')
        time.sleep(2)
        connection_error = False
        try:
            pg = create_engine(f'postgresql://postgres:{postgres_pwd}@0.0.0.0:5555/postgres')     # try to connect to pg
        except:
            print('Connection failed. Trying again in 10 seconds')
            time.sleep(10)
            try:
                pg = create_engine(f'postgresql://postgres:{postgres_pwd}@0.0.0.0:5555/postgres') # try to connect to pg
            except:
                connection_error = True
        if connection_error:
            print('Connection failed. You can try again later using the csvimport.py script in the /config folder.')
        else:
            tracksimport = pd.read_csv('config/tracks.csv', delimiter = ',', decimal = '.')
            tracksimport.iloc[:,:91] = tracksimport.iloc[:,:91].apply(pd.to_numeric, errors='coerce', downcast='float')
            tracksimport[['trackid', 'year', 'labelid']] = tracksimport[['trackid', 'year', 'labelid']].apply(pd.to_numeric, errors='coerce', downcast='integer')
            tracksimport.to_sql('tracks', pg, if_exists='replace', method='multi', index = False, chunksize=1000)  # import to postgres
            labelsimport = pd.read_csv('config/labels/labels.csv', delimiter = ',', decimal = '.')
            labelsimport[['labelid', 'numtracks']] = labelsimport[['labelid', 'numtracks']].apply(pd.to_numeric, errors='coerce', downcast='integer')
            labelsimport.to_sql('labels', pg, if_exists='replace', method='multi', index = False, chunksize=1000)  # import to postgres
            pg.dispose()
            print("Database import complete.")
    else:
        print('Error: could not find one or any of these files: config/tracks.csv; config/labels/labels.csv')
    

def main():
    """
    Console script main function. Options: -on -off

    """    
    if len(sys.argv)==1:    
        print('Bandcamp Dance Librarian console script. Options: -on -off')
    else:
        parser = argparse.ArgumentParser(description='Bandcamp Dance Librarian console script. Options: -on -off')
        parser.add_argument('-on', help='Start the application Docker containers. If no environmental variables file found, it initializes one and builds the containers.', action='store_true')
        parser.add_argument('-off', help='Stops the application Docker containers.', action='store_true')
        args, unknown = parser.parse_known_args()
        if len(unknown) > 0:
            print(f'Undefined: {unknown[0]}')
        elif args.on:
            if args.off:
                print('Cannot use -on and -off at the same time.')
            else:
                os.chdir(os.path.dirname(os.path.realpath(__file__)))
                if os.path.exists('.vars.env'):
                    print('If requested, please enter your root password for starting the Bandcamp Dance Librarian Docker containers in the background.')
                    try:
                        output = subprocess.check_output(['bash','-c', 'sudo docker-compose kill'])
                        output = subprocess.check_output(['bash','-c', 'sudo docker-compose up --detach --force-recreate'])
                    except:
                        output = subprocess.check_output(['bash','-c', 'docker-compose kill'])
                        output = subprocess.check_output(['bash','-c', 'docker-compose up --detach --force-recreate'])     # for AWS etc. deployment
                    print('The web interface should now be running on http://0.0.0.0:8000/. For stopping the docker service from the command line, please enter: bandcamplibrarian -off')
                else:
                    print('Please note that Docker Compose needs to be installed on this machine before deployment!')
                    print('Could not find file containing environmental variables. Initializing and building Docker containers...')
                    postgres_pwd = input('Please enter a password for the Postgres database: ')
                    f = open('.vars.env', 'w')
                    f.write('POSTGRES_USER=postgres')
                    f.write('\n')
                    f.write(f'POSTGRES_PASSWORD={postgres_pwd}')
                    f.write('\n')
                    f.write('POSTGRES_DB=postgres')
                    f.write('\n')
                    flask_key=os.urandom(32).hex()
                    f.write(f'FLASK_KEY={flask_key}')                    
                    f.close()
                    print('If requested, please enter your root password for deploying the Docker services of Bandcamp Dance Librarian.')
                    try:
                        output = subprocess.check_output(['bash','-c', 'sudo docker-compose kill'])
                        output = subprocess.check_output(['bash','-c', 'sudo docker-compose rm -fv pg_container'])
                        output = subprocess.check_output(['bash','-c', 'sudo docker-compose up --detach --build --force-recreate'])
                    except:
                        output = subprocess.check_output(['bash','-c', 'docker-compose kill'])
                        output = subprocess.check_output(['bash','-c', 'docker-compose rm -fv pg_container'])
                        output = subprocess.check_output(['bash','-c', 'docker-compose up --detach --build --force-recreate'])     # for AWS deployment
                    if 'Successfully built' in output.decode("utf-8"):
                        print('The web interface should now be running on http://0.0.0.0:8000/.')
                        choice = 'x'
                        while choice not in ['y','Y','n','N']:
                            choice = input('Would you like to add the samples database including 10 labels and 6486 tracks? (y/n) ')
                        if choice in ['y','Y']:
                            import_tracks(postgres_pwd)
                        print('Please refer to the documentation/readme for further app settings (accessible via config/config.csv). For stopping the docker service from the command line, use: bandcamplibrarian -off')
                    else:
                        print('Error detected! Please refer to the Docker documentation and deploy the containers manually.')
        elif args.off:
            os.chdir(os.path.dirname(os.path.realpath(__file__)))
            print('If requested, please enter your root password for stopping the docker service of the Bandcamp Dance Librarian.')
            try:
                output = subprocess.check_output(['bash','-c', 'sudo docker-compose kill'])
            except:
                output = subprocess.check_output(['bash','-c', 'docker-compose kill'])


if __name__ == "__main__":
    main()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Exports the tracks table into a .cvs file
"""

from sqlalchemy import create_engine
import pandas as pd
import psycopg2

print("Please note that you need a running instance of the Postgres docker container for the CSV export")
pguser = input("User name:")
pgpassword = input("Password:")
pg = create_engine(f'postgresql://{pguser}:{pgpassword}@0.0.0.0:5555/postgres') # pg connect
tracks = pd.read_sql_table('tracks', pg)
tracks.to_csv('tracks.csv', index=False)
print("Table exported to tracks.csv")
pg.dispose()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

Builds classification model using the Beatport 2021 Jan dataset
Follows the methodology of Caparrini et al. (https://doi.org/10.1080/09298215.2020.1761399)
Uses 92 features extracted with pyAudioAnalysis and Essentia

"""

import pickle
import pandas as pd

from sklearn.ensemble import ExtraTreesClassifier
from sklearn.model_selection import train_test_split, cross_val_score
from sklearn.metrics import f1_score


print('Generating model..')
df = pd.read_csv('beatport_2021jan.csv')
X = df.drop(['genre','artist','track'], axis=1)
y = df['genre']
X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, stratify=y, random_state=42)

classweight = {"Afro House": 1,"Ambient": 0.5,"Bass House": 1,"Big Room": 1,"Breaks": 1,"Dance-Electro Pop": 1,"Deep House": 1,"Drum&Bass": 1,"Dubstep": 1,"Electro (Classic-Detroit-Modern)": 1,"Electro House": 1,"Funky-Groove-Jackin' House": 1,"Future House": 1,"Garage-Bassline-Grime": 1,"Hard Dance-Hardcore": 1,"Hard Techno": 1,"House": 1,"Indie Dance": 1,"Leftfield Bass": 1,"Leftfield House & Techno": 1,"Melodic House&Techno": 1,"Minimal-Deep Tech": 1,"Nu Disco-Disco": 1,"Organic House-Downtempo": 1,"Progressive House": 1,"Psy-Trance": 1,"Reggae-Dancehall-Dub": 1,"Tech House": 1,"Techno (Peak Time-Driving)": 1,"Techno (Raw-Deep-Hypnotic)": 1,"Trance": 1,"Trap-Hip-Hop-R&B": 1}    
BP2021 = ExtraTreesClassifier(bootstrap=False, class_weight=classweight, criterion='gini',
           max_depth=None, max_features=40,
           max_leaf_nodes=None, min_impurity_decrease=0,
           min_samples_leaf=1, min_samples_split=12,
           min_weight_fraction_leaf=0, n_estimators=125, n_jobs=4,
           oob_score=False, random_state=42, verbose=0, warm_start=False)

accuracy = cross_val_score(BP2021, X_train, y_train, cv=10)
print(f'Train CV (10-fold) accuracy: {round(accuracy.mean(),3)} +/- {round(accuracy.std(), 3)}')
BP2021.fit(X_train, y_train)
print(f'Test F1 score: {round(f1_score(y_test, BP2021.predict(X_test), average="macro"), 3)}')
print('Dumping classifier into the /config folder.')
pickle.dump(BP2021, open('../config/beatport_classifier.sav', 'wb'))


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
        'numpy',
        'pandas',
        'sqlalchemy',
        'psycopg2-binary',]

setup(
    author="Botond Vitos",
    author_email='boti@vitos.tv',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    description="This project detects stylistic tendencies in the Bandcamp libraries of electronic dance music labels by defining clusters based on Beatport's subgenre categories as of Jan 2021. To deploy the web interface on 0.0.0.0:8000/, enter bandcamplibrarian -on from the command line after installation. Note: the deployment requires Docker Compose pre-installed on your system.",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='bandcamp_librarian',
    name='bandcamp_librarian',
    packages=find_packages(include=['bandcamp_librarian','bandcamp_librarian.*']),
    url='https://github.com/bvitos/bandcamp_librarian',
    version='0.2.2',
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'bandcamplibrarian=bandcamp_librarian.cli:main',
        ],
    }
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""

Estimates clusters among tracks of Bandcamp labels, 
based on features extracted with pyAudioAnalysis and Essentia

The classification model is based on the Electronic Music Features - 201802 BeatportTop100 dataset
https://www.kaggle.com/caparrini/electronic-music-features-201802-beatporttop100
See Caparrini et al. 2020 (https://doi.org/10.1080/09298215.2020.1761399)

"""

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import FuncFormatter
import seaborn as sns

import pickle
import os
import gc
import logging
import spacy
import requests

#from sklearn.cluster import KMeans
from sklearn.cluster import MiniBatchKMeans
from scipy.spatial.distance import cdist
from kneed import KneeLocator

from PIL import Image
from fpdf import FPDF
from sqlalchemy import create_engine
from random import randrange
from collections import Counter


def clusterise_label(df, labelname, n_clusters, prediction_weight, randomstring):
    """
    Distribute label stored in input dataframe into n clusters

    Returns:
        array of strings used for generating the results PDF file

    """
    clf = pickle.load(open('../config/beatport_classifier.sav', 'rb')) 
    
    features_table = df.iloc[:,:92].copy(deep=True)
    proba_features=pd.DataFrame(data=clf.predict_proba(features_table.to_numpy()),columns=clf.classes_)          # predicted features df -> stores subgenre probabilities for each track

    attempts = 1
    while attempts > 0:                                                 # attempts will be set to 0 once elbow point is found during automatic cluster number detection OR if cluster number is user-defined

        inertia = []
        if n_clusters == 0:                                             # automatic cluster number detection: the loop will calculate 6 inertia values and search for the elbow point of the curve
            K = range(1, 7)
        else:
            K = range(n_clusters, n_clusters + 1)                       # user-defined cluster numbers: the loop will only run once
        Kmean_version = []                                              # stores the K-means models for each version of cluster numbers during automatic detection
        for cluster_nr in K:
            confidence_max = 0
            for i in range(20):                                         # run the K-means algorithm multiple times and store the results with the highest cumulative confidence values of the cluster centres provided by the subgenre classificator
                Kmean_try = MiniBatchKMeans(init="k-means++", n_clusters=cluster_nr) # this way, clusters that are crystallised around the existing taxonomy should be selected // MiniBatchKmeans for faster results
#                Kmean_try = KMeans(init="k-means++", n_clusters=cluster_nr)
                Kmean_try.fit(proba_features)
                confidence_sum = 0
                clustercenters = Kmean_try.cluster_centers_             # store cluster centers
                for j in range(cluster_nr):
                    element = clustercenters.tolist()[j]                # element = the centroid probability values of each cluster
                    pred_list = list(zip(clf.classes_, element))
                    pred_list.sort(key = lambda x: x[1], reverse=True)  # sort them in order of probability
                    confidence_sum += pred_list[0][1] * prediction_weight + pred_list[1][1] + pred_list[2][1] # add together the highest probability values, taking into account the cluster genre prediction weight in order to gain purer subgenres (2x) or a better amalgam of subgenres (0.5x)
                if confidence_sum > confidence_max:                     # store the model with the highest cumulative confidence value
                    confidence_max = confidence_sum
                    Kmean = Kmean_try
            Kmean_version.append(Kmean)
            inertia.append(Kmean.inertia_)

        if n_clusters == 0:                                             # in case of automatic cluster numbers detection: determine the elbow point of the inertia curve, and store it as n_clusters
            kl = KneeLocator(K, inertia, curve="convex", direction="decreasing")
            n_clusters = kl.elbow
            logging.info(f'Number of clusters for {labelname}: {n_clusters}.')
            try:
                Kmean = Kmean_version[n_clusters - 1]                   # if valid elbow point found, store the final model
            except:
                if attempts < 5:                                        # if no elbow point found (n_clusters is None): repeat the loop max. 5 times
                    attempts += 1
                    n_clusters = 0
                else:
                    attempts = 0
                    n_clusters = 4                                      # if elbow point cannot be determined after 5 attempts, just use a default nr of 4 clusters
                    Kmean = Kmean_version[n_clusters - 1]               # store the final model
        else:
            attempts = 0                                                # if cluster number was defined by user, only run the loop once

    clusters_pred = Kmean.predict(proba_features)                       # clusters_pred will be the Cluster column of the main df
    df['Cluster']=np.array(clusters_pred)
    proba_features['Cluster']=np.array(clusters_pred)                   # as well as the proba_features df
    clustertext = []
    clustercenters = Kmean.cluster_centers_                             # get the  cluster centers
    
    for ccounter in range(n_clusters):
        np.set_printoptions(formatter={'float': lambda x: "{0:0.4f}".format(x)})

        element = clustercenters.tolist()[ccounter]                     # element = the centroid probability values of each cluster
        pred_list = list(zip(clf.classes_, element))
        pred_list.sort(key = lambda x: x[1], reverse=True)              # sort it in order of probability, the subgenres with the three highest values will be stored:
        cluster_df = df[df['Cluster']==ccounter].copy(deep=True)        # store the dataframe cluster in a separate view for easier reference
        cluster_features = proba_features[proba_features['Cluster']==ccounter].copy(deep=True)     # store the subgenre probas dataframe cluster in a separate df view for easier reference
        clustertext.append(f'Cluster {ccounter}- elements of {pred_list[0][0]} ({round(pred_list[0][1],2)}), {pred_list[1][0]} ({round(pred_list[1][1],2)}) and {pred_list[2][0]} ({round(pred_list[2][1],2)}).')
        locations = []
        for row in cluster_df['GPE']:
            locations = locations + row
        nationalities = []
        for row in cluster_df['NORP']:
            nationalities = nationalities + row
        locationcount = Counter(locations)
        nationalitiescount = Counter(nationalities)
        loctags = 'Locale tags'
        for locelement in sorted(locationcount, key=locationcount.get, reverse=True):
            loctags = loctags + '- ' + locelement + ' (' + str(locationcount[locelement]) + ') '
        for natelement in sorted(nationalitiescount, key=nationalitiescount.get, reverse=True):
            loctags = loctags + '- ' + natelement + ' (' + str(nationalitiescount[natelement]) + ') '
        clustertext.append(loctags)
        clustertext.append(f"The cluster includes {cluster_df.shape[0]} tracks. Notable track examples (clicking on the cover will take you to the Bandcamp release page):")

        distances = cdist(cluster_features.iloc[:,:32], np.array(element).reshape(1,-1), 'euclidean').flatten().tolist() # calculate the distances from subgenre probabilities centroid

        asc_index = np.argsort(distances).tolist()                      # get the indexes of sorted distances from centroid
        found_duplicate = True
        while (len(asc_index) > 3) & (found_duplicate):                 # try to select three tracks from different artists
            found_duplicate = False
            for j in range(3):
                artist_list = [cluster_df.iloc[asc_index[0],93], cluster_df.iloc[asc_index[1],93], cluster_df.iloc[asc_index[2],93]]
                artist_list.pop(j)
                if (cluster_df.iloc[asc_index[j],93] in artist_list) & (cluster_df.iloc[asc_index[j],93] != 'Various') & (cluster_df.iloc[asc_index[j],93] != 'Various Artists'):
                    duplicate = asc_index[j]
                    found_duplicate = True
            if (found_duplicate):
                asc_index.remove(duplicate)
        j = 0
        while j < 3 and len(asc_index) > j:                             # stores the three most fitting (closest to their centroid) track examples (provided that they exist) with links and release covers
            clustertext.append(f'Link: {cluster_df.iloc[asc_index[j],100]}')
            clustertext.append(cluster_df.iloc[asc_index[j],101])
            imgstr = cluster_df.iloc[asc_index[j],103]
            if len(imgstr) > 0:
                try:                                                    # try to download cover image
                    image_data = requests.get(imgstr).content
                    filename = f'../config/labels/{randomstring}{cluster_df.iloc[asc_index[j],92]}{str(ccounter * 3 + j)}.jpg'
                    f = open(filename, 'w+b')
                    f.write(image_data)
                    f.close()
                    img = Image.open(filename)
                    img = img.resize((300,300), Image.ANTIALIAS)
                    img.save(filename)
                    clustertext.append(filename)
                except:                                                 # no cover image available
                    clustertext.append('../config/_nocover_.jpg')
            else:
                clustertext.append('../config/_nocover_.jpg')
            clustertext.append(cluster_df.iloc[asc_index[j],97])
            j += 1
        clustertext.append("Note: the track analysis and consecutive K-means clustering is based on Beatport's taxonomy composed of 32 subgenres as of Jan 2021.")
        clustertext.append("The classifier was trained on the top 100 tracks of each subgenre; the numbers in brackets represent probability/confidence values.")
    fig = plt.figure()
    sns.histplot(data=df, x="year", hue="Cluster", palette = "Set2", element = "poly")
#   sns.kdeplot(data=df, x="Year", hue="Cluster", shade=True, palette = "Set2")             # alternative to histplot
    plt.gca().xaxis.set_major_formatter(FuncFormatter(lambda x, _: int(x)))                 # only display integers on X axis
    plt.title(df['labelname'].values[0])
    plt.xlabel('Year')
    plt.ylabel('Tracks')
    fig.savefig(f'../config/labels/{randomstring}{labelname}.png', dpi=200)
    plt.close()

    return clustertext



def extract_locales(tags, blacklist=list, GPE_whitelist=list, locale_type=str):
    """
    Extracts locale information from a string containing tags separated by '|' while disregarding blacklisted tags
    spaCy entity types: GPE (Geopolitical entity); NORP (Nationalities or religious or political groups)

    Returns:
        list of locations

    """
    if locale_type+tags in tags_cache:                  # check if the tag/entity pair was already processed; duplicates are common
        return tags_cache[locale_type+tags]
    else:
        results = []
        for tag in tags.split('|'):
            if locale_type == 'GPE' and tag in GPE_whitelist and tag not in results: # add whitelisted location tags instantly
                results.append(tag)
            for ent in nlp(tag.title()).ents:           # check for location info in individual tag (in title case):
                if ent.label_ in [locale_type] and ent.text not in results and ent.text not in blacklist:
                    results.append(ent.text)
        tags_cache[locale_type+tags] = results          # store processed tag/entity string in cache dictionary
        return(results)





def process_label(labelname, clusternum):
    """
    Clusterise/classify a Bandcamp library

    Returns:
        the function does not return anything
    Instead, it generates a PDF file containing the classification results (clusters, subgenres, Bandcamp track examples incl. covers, tags/folksonomies and links)

    """
    labels = [labelname]                                                            # the script currently processes one label/library only, but with an easy modification of the code multiple labels can be processed at the same time
    pguser = os.environ.get('POSTGRES_USER')                                        # setup postgres connection
    pgpassword = os.environ.get('POSTGRES_PASSWORD')
    pg = create_engine(f'postgresql://{pguser}:{pgpassword}@pg_container:5432/postgres')
    cltext = []
    if pg.dialect.has_table(pg, 'tracks'):
        randomstring = str(randrange(100000000,999999999))
        df_full = pd.read_sql_table('tracks', pg)
        df = df_full[df_full['bclabel']==labelname].copy(deep=True)
        configdata=pd.read_csv('../config/config.csv')
        loc_tag_blacklist = configdata['loc_tag_blacklist'].values[0].split('|')   # tags incorrectly detected as locales by the NLP algorithm must be blacklisted
        GPE_tag_whitelist = configdata['loc_tag_whitelist'].values[0].split('|')   # locale tags not detected by the NLP algorithm must be whitelisted
        prediction_weight = configdata['prediction_weight'].values[0]              # if the value is 2, the cluster/genre prediction will be aimed towards purer subgenres; if it is 0.5, it will be aimed towards a better amalgam of subgenres
        global nlp
        nlp = spacy.load('en_core_web_md')
        global tags_cache
        tags_cache = {}
        df['GPE'] = df['tags'].apply(extract_locales, args=[loc_tag_blacklist, GPE_tag_whitelist, 'GPE'])       # extract location info into df column
        df['NORP'] = df['tags'].apply(extract_locales, args=[loc_tag_blacklist,GPE_tag_whitelist, 'NORP'])      # extract location info into df column
        cltext.append(clusterise_label(df, labelname, int(clusternum), prediction_weight, randomstring))
    else:
        cltext.append(['Error: Tracks table not found in the database.'])
    pg.dispose()        
    pdf = FPDF()
    for i in range(len(labels)):                                                    # generate PDF report
        pdf.add_page()
        pdf.set_auto_page_break(0, margin = 0.5)
        labeltext = cltext[i]
        labeltext = [x.encode('latin-1', 'replace').decode('latin-1') for x in labeltext]
        labeltext = [x.replace('&amp;', '&') for x in labeltext]
        for j in range(len(labeltext)):
            pdf.add_font('DejaVu', '', '../config/DejaVuSansCondensed.ttf', uni=True)
            pdf.add_font('DejaVuI', '', '../config/DejaVuSansCondensed-Oblique.ttf', uni=True)
            pdf.set_font('DejaVu', size=8)
            text2 = labeltext[j]
            if text2[0:5] == 'Clust':
                if j > 0:
                    pdf.add_page()
                pdf.image(f'../config/labels/{randomstring}{labels[i]}.png', x=30, y=0, w=140)        
                pdf.set_y(110)
                pdf.set_font('DejaVu', size=9)
                pdf.cell(200, 9, txt=text2, ln=1, align="C")
                pdf.set_font('DejaVu', size=9)
                pdf.cell(200, 9, txt=labeltext[j+1], ln=20, align="C")
                pdf.cell(200, 9, txt=labeltext[j+2], ln=20, align="C")
            elif text2[0:5] == 'Link:':
                top = pdf.y
                pdf.image(labeltext[j+2], x=5, y=pdf.y, w=45, h=45, link = labeltext[j+1])
                if labeltext[j+2] != '../config/_nocover_.jpg':
                    os.remove(labeltext[j+2])
                pdf.y = top
                pdf.x = 50
                pdf.set_font("DejaVu", size=7)
                celltext = f"{text2[text2.find('_') + 1:-4]} \n Bandcamp artist tags (folskonomy): \n {labeltext[j+3]}"
                pdf.multi_cell(155, 15, txt=celltext, border=1)
                pdf.y = top + 45
                pdf.x = 0
            elif text2[0:5] == 'Note:':
                pdf.set_font('DejaVuI', size=8)
                pdf.x = 0
                pdf.y = 275
                pdf.cell(200, 8, txt=text2, align="C")
                pdf.x = 0
                pdf.y = 280
                pdf.cell(200, 8, txt=labeltext[j+1], align="C")
            elif text2[0:5] == 'Error':
                pdf.set_font('DejaVuI', size=10)
                pdf.y = 140
                pdf.cell(200, 12, txt=text2, align="C")
    os.remove(f'../config/labels/{randomstring}{labels[i]}.png')
    pdf.output(f'../config/labels/{labelname}{clusternum}.pdf')
    gc.collect()
